# $id $

# BBNorm.pm

=head1 NAME

    BBNorm.pm

=cut

=head1 SYNOPSIS

    my $bbnorm_obj = new BBNorm( \%astral_id_weight, $fam_norm_lite_object);
    print $fam_norm_obj->normFactor($ID1, $ID2);

    %astral_id_weight is a hash with the astral db ids as keys, and their 
    bayesian bootstrap weights as the values.
    
    $fam_norm_lite_object is a reference to the FamNormLite object initialized 
    with the astral db..

=cut

=head1 DESCRIPTION 

    BBNorm makes the necessary db hashes to perform normalization for any
    two astral IDs, and then provides normalization arrays in the form:
    [no norm, lin norm, quad norm] on the fly, given two ids. The constructor
    makes 2 hashes: one with all the superfamilies in the astral database as
    keys and all the ids in each superfamily as values, and one with
    the superfamily names as keys and an array containing: [coverage size, linear
    coverage size, # of sequence entries] for that superfamily.

    If w(i) is the weight of any particular astral id in a given superfamily and
    i = 1 - superfamily size, then the family coverage size (CS) is sum(w(i)*w(j)) 
    where i != j, and the linear coverage size (LCS) is simply the sum of the 
    weights of each astral id in the superfamily.

    normFactor() performs on demand normalization for any two superfamily ids.
    It returns the array [standard weight, linear weight, quadratic weight] where
    standard weight (SW) is simply w1 * w2, linear weight is SW/(LCS - w1), and 
    the quadratic weight is SW/CS. If the ids are not related the linear and
    quadratic weights are 0.

    covSizes() returns the total coverage size for the astral database - it merely
    sums the coverage (SCS) and linear coverage sizes (SLCS) for the superfamilies,
    and returns the array [SCS, SLCS, # of superfamilies].


=cut

=head1 AUTHOR

    (c) 2002 Gavin Price, UC Regents

=cut 

package BBNorm;
use strict;

=head2 new

  Title:    new
  Function: constructor
  Returns:  a new BBNorm object
  Args:     a hash of sequence IDs and their weights, a ref to a FamNormLite object

=cut

sub new {
    my ( $class, $astral_ids, $fnlo) = @_;
    my $self = &_makeSFDB( $astral_ids, $fnlo);
    $self->{'astral_ids'} = $astral_ids;
    $self->{'fnlo'} = $fnlo;
    bless( $self, $class );
    return $self;
}

=head2 _makeSFDB

  Title:    _makeSFDB
  Function: internal
  Returns:  object data structure
  Args:     a hash of sequence IDs and their weights, a ref to a FamNormLite object

=cut

sub _makeSFDB{
    my %astral_ids = %{shift(@_)};
    my $fnlo = shift;
    my (%SFs, %self, %SFcontents, %SFnames);
    my (@iDarray);
    my ($iD, $iD1, $iD2, $superFamily, $famCov, $famSize, $famNumEntries);
    my ($c1, $c2, $c3, $c4, $c5);
    

    #make hash with list of all IDs in a superfamily for each superfamily
    for $iD (keys %astral_ids) {
	if ($SFcontents{ $fnlo->idToSF($iD) } == undef) {  #if value empty
	    $SFcontents{ $fnlo->idToSF($iD) } = [];        #init it
	}
	unshift(@{$SFcontents{ $fnlo->idToSF($iD) }}, $iD);
    }		    

    #make hash of family coverage sizes
    for $superFamily (keys %SFcontents) {       #run through each superfamily
	@iDarray = @{$SFcontents{ $superFamily }};
	$famCov = 0; $famSize = 0; $famNumEntries = 0;
	for $iD1 (@iDarray) {                   #run through every ID in that SF
	    for $iD2 (@iDarray) {               #compare it against each ID
		if ($iD1 ne $iD2) {
		    $famCov += $astral_ids{$iD1} * $astral_ids{$iD2}; #coverage size
		}
	    }
	    $famSize += $astral_ids{$iD1};                            #linear coverage size
	    $famNumEntries++;
	}
	$SFs{$superFamily} = [$famCov, $famSize, $famNumEntries];   #store coverage array
    }

    $self{'SFs'} = \%SFs;		    
    $self{'SFcontents'} = \%SFcontents;
    return \%self;
}

=head2 normFactor

  Title:    normFactor
  Function: accessor
  Returns:  array of normalization factors: [Std wgt, lin wgt, quad wgt]
  Args:     two astral IDs

=cut

sub normFactor {
    my $self = shift;
    my $iD1 = shift;
    my $iD2 = shift;
    my ($stdWgt, $linWgt, $quadWgt);
    my ($condition1, $condition2, $condition3);    

    my $astral_ids = $self->{'astral_ids'};
    my $SFs = $self->{'SFs'};
    my $fnlo = $self->{'fnlo'};

    $stdWgt = $astral_ids->{$iD1} * $astral_ids->{$iD2};          #calc std wgt
    my @covArray = @{$SFs->{ $fnlo->idToSF($iD1) }};            #get SF cov array
    $condition1 = $covArray[2] > 1;                     #SF has more than 1 entry
    $condition2 = $fnlo->idToSF($iD1) eq $fnlo->idToSF($iD2); #ids in same SF
    $condition3 = $iD1 ne $iD2;                         #non duplicate ids
    if ($condition1 and $condition2 and $condition3) {
	$linWgt = $stdWgt/($covArray[1] - $astral_ids->{$iD1});  #calc lin wgt
	$quadWgt = $stdWgt/$covArray[0];                       #calc quad wgt
    } else {
	$linWgt = 0;
	$quadWgt = 0;
    }
    my @return_array = ($stdWgt, $linWgt, $quadWgt);
    return \@return_array;
}

=head2 covSizes

  Title:    covSizes
  Function: accessor
  Returns:  array of DB coverage sizes: [coverage, linear coverage, quad coverage]
  Args:     none

=cut

sub covSizes{
    my $self = shift;
    my $superfamily;
    
    if ( $self->{'covSizes'} ) {         #if we've already done this
	return @{ $self->{'covSizes'}};  #just send the prev. results
    } else {
	my %SFs = %{$self->{'SFs'}};
	my @covSizes = (0,0,0);
	for $superfamily (keys(%SFs)) {  #run through each superfamily
	    if ($SFs{$superfamily}->[2] > 1) {     #if there's more than one sequence
		$covSizes[0] += $SFs{$superfamily}->[0];  #sum each type of coverage
		$covSizes[1] += $SFs{$superfamily}->[1];
		$covSizes[2] += 1;
	    }
	}
	$self->{'covSizes'} = \@covSizes;
	return @covSizes;
    }
}
1;
