# Module for parsing /accessing dbvdb output
# Ed Green
# (c) 2002 / Ed Green, UC Berkeley

=head1 NAME

DBVDB.pm

=head1 SYNOPSIS

my $dbvdb_object = new DBVDB( 'dbvdb.res' );

print $dbvdb_object->numHits();

foreach my $query ( @{ $dbvdb_object->queryIds() } ) {

    foreach my $hit ( @{ $dbvdb_object->bestNonOverQueHits( $query ) } ) {

	# do something with the hit

	# it's an array like this

	# ( query_id

	#   target_id 

	#   significance score, 

	#   scop relatedness, 

	#   unnorm weight, 

	#   lin. norm. weight, 

	#   quad. norm. weight,
	
    }

}

=head1 DESCRIPTION

DBVDB is a module that parses a dbvdb output file, stores the
data internally and makes the data available through the methods
described below

=cut

package DBVDB;
use strict;

sub new {
    my ( $class, $file ) = @_;
    my $self = &_parse( $file );
    bless( $self );
    return $self;
}

sub sigs {
    my $self = shift;
    my @sigs = sort { $a <=> $b } keys %{ $self->{ 'hits' } };
    return @sigs;
}

sub hitsAtSig {
    my $self = shift;
    my $sig = shift;
    my $hits = $self->{ 'hits' }->{ $sig };
    return $hits;
}

sub queries {
    my $self = shift;
    return keys %{ $self->{ 'query_hits' } };
}

sub targets {
    my $self = shift;
    return keys %{ $self->{ 'target_hits' } };
}

sub queryDb {
    my $self = shift;
    return $self->{ 'qdb' };
}

sub targetDb {
    my $self = shift;
    return $self->{ 'tdb' };
}

sub numQueries {
    my $self = shift;
    return $self->{ 'num_q' };
}

sub numTargets {
    my $self = shift;
    return $self->{ 'num_t' };
}

sub numQueryRels {
    my $self = shift;
    return $self->{ 'num_q_rels' };
}

sub numTargetRels {
    my $self = shift;
    return $self->{ 'num_t_rels' };
}

sub hitQueryId {
    my $self  = shift;
    my $hit_p = shift;
    return $hit_p->[ 0 ];
}

sub hitTargetId {
    my $self  = shift;
    my $hit_p = shift;
    return $hit_p->[ 1 ];
}

sub hitSig {
    my $self = shift;
    my $hit_p = shift;
    return $hit_p->[ 2 ];
}

sub hitRel {
    my $self = shift;
    my $hit_p = shift;
    return $hit_p->[ 3 ];
}

sub hitWeightUnnorm {
    my $self = shift;
    my $hit_p = shift;
    return $hit_p->[ 4 ];
}

sub hitWeightLinNorm {
    my $self = shift;
    my $hit_p = shift;
    return $hit_p->[ 5 ];
}

sub hitWeightQuadNorm {
    my $self = shift;
    my $hit_p = shift;
    return $hit_p->[ 6 ];
}

sub _parse {
    my $fn = shift;
    my ( %hits, # keyed by sig
	 %queries, # keyed by id to array of hits
	 %targets, # keyed by id to array of hits
	 %self,
	 $l );

    open( FH, $fn ) or die ( "$fn: $!" );
    $l = <FH>; &_checkHead( $l );
    $l = <FH>; my $qdb = &_getDb( $l );
    $l = <FH>; my $tdb = &_getDb( $l );
    $l = <FH>; my $num_q = &_getNum( $l );
    $l = <FH>; my $num_t = &_getNum( $l );
    $l = <FH>; my @num_q_rels = &_getRels( $l );
    $l = <FH>; my @num_t_rels = &_getRels( $l );
    <FH>;
    <FH>;
    while ( my $line = <FH> ) {
	my @this_hit = &_parseHit( $line );
	push( @{ $hits{ $this_hit[ 2 ] } }, \@this_hit );
	push( @{ $queries{ $this_hit[ 0 ] } }, \@this_hit );
	push( @{ $targets{ $this_hit[ 1 ] } }, \@this_hit );
    }
    $self{ 'qdb' } = $qdb;
    $self{ 'tdb' } = $tdb;
    $self{ 'num_q' } = $num_q;
    $self{ 'num_t' } = $num_t;
    $self{ 'num_q_rels' } = \@num_q_rels;
    $self{ 'num_t_rels' } = \@num_t_rels;
    $self{ 'hits' } = \%hits;
    $self{ 'query_hits' } = \%queries;
    $self{ 'target_hits' } = \%targets;
    return \%self;
}

sub _checkHead {
    my $head = shift;
    unless( $head =~ /dbvdbsum v\d+\.\d+ results/ ) {
	die( "doesn't look like a dbvdb output file to me" );
    }
}

sub _getDb {
    my $db_line = shift;
    my ($db) = ($db_line =~ /\S+ database: (.+)/);
    return $db;
}

sub _getNum {
    my $num_line = shift;
    my ($num) = ($num_line =~ /\S+ database size: (\d+)/);
    return $num;
}

sub _getRels {
    my $rel_line = shift;
    my (@rel) = split( /\s+/, $rel_line );
    return ( @rel[ 3..5 ] );
}

sub _parseHit {
    my $hit_line = shift;
    my (@el) =
	($hit_line =~/(\S+) (\S+) (\S+) (\S+) (\S+) (\S+) (\S+)/);
    return @el;
}

1;
