# $id $

# (c) 2002 Ed Green, UC Regents

# FamNorm.pm

=head1 NAME

    FamNorm.pm

=cut

=head1 SYNOPSIS

    my $fam_norm_obj = new FamNorm( 'astral_file' );
    print $fam_norm_obj->totalSFs();

=cut

=head1 DESCRIPTION 

    FamNorm is a simple module which reads a sasum database file and
    stores information about the superfamilies contained within it.
    Methods are available for extracting this information.
    
    This module is particularly useful for doing family normalization
    of sequence comparison results.  There is a convention which runs
    through the module which is that there are 3 ways to do family
    normailzation.  The first, is no family normalization.  The second
    is to scale each entry by 1/(n*n - n) where n is the size of the
    family.  This means that all families are weighted equally,
    regardless of size.  The third way is to scale each entry by
    1/(n-1).  This means that each superfamily is weighted linearly by
    its number of entries.  The methods normFactor() and covSizes()
    return 3 element array which are in the order just described.

=cut

=head1 AUTHOR

    (c) 2002 Ed Green, UC Regents

=cut 

package FamNorm;
use strict;

=head2 new

  Title:    new
  Function: constructor
  Returns:  a new FamNorm object
  Args:     an astral database

=cut

sub new {
    my ( $class, $astral_file, $rest_file ) = @_;
    my $self = &_parseInput( $astral_file, $rest_file );
    bless( $self );
    return $self;
}

=head2 astralIds

  Title:    astralIds
  Function: accessor
  Returns:  array of astral ids for this set; if a sasum file was
    specified to the constructor, then the ids returned here will be
    limited to those seen as queries in that sasum file
  Args:     none

=cut

sub astralIds {
    my $self = shift;
    my @ids = keys %{ $self->{ 'astral_ids' } };
    return @ids;
} 

=head2 sfIds

  Title:    sfIds
  Function: accessor
  Returns:  array of superfamily ids
  Args:     none

=cut

sub sfIds {
    my $self = shift;
    my @ids = keys %{ $self->{ 'SFs' } };
    return @ids;
}

=head2 numSFEntries

  Title:    numSFEntries
  Function: accessor
  Returns:  scalar size of this family (number of members), some of
    which may be repeated
  Args:     family id (really superfamily id -- like 'a.1.1')

=cut

sub numSFEntries {
    my $self = shift;
    my $sf   = shift;
    return $self->{ 'SFs' }->{ $sf }->[ 0 ];
}

=head2 sfHitSpace

  Title:    sfHitSpace
  Function: accessor
  Returns:  number of correct hits that could be found in this sf
    n^2 - SUM( m^2 ), where m is the number of occurances of
    each unique member
  Args:     family id (really superfamily id -- like 'a.1.1')

=cut

sub sfHitSpace {
    my $self = shift;
    my $sf   = shift;
    return $self->{ 'SFs' }->{ $sf }->[ 1 ];
}

=head2 numSFUniq

  Title:    numSFUniq
  Function: accessor
  Returns:  number of unique entries in this superfamily
  Args:     family id (really superfamily id -- like 'a.1.1')

=cut

sub numSFUniq {
    my $self = shift;
    my $sf   = shift;
    return $self->{ 'SFs' }->{ $sf }->[ 2 ];
}

=head2 normFactor

  Title:    normFactor
  Function: accessor
  Returns:  3 element array of normalization factors for this family
  Args:     family id (really superfamily id -- like 'a.1.1')

=cut

sub normFactor {
    my $self = shift;
    my $sf   = shift;
    my $numSFUniq = $self->numSFUniq( $sf );
    if ( $numSFUniq == 1 ) {
	# if there's only 1 unique sequence, we don't get credit for
	# that so it does not really matter what factors are returned
	return ( 1, 1, 1 );
    }
    elsif ( $self->sfHitSpace( $sf ) == 0 ) {
	return ( 1,
		 ( 1 / $self->sfHitSpace( $sf ) ),
		 0 );
    }
    else {
	return ( 1,
		 ( 1 / $self->sfHitSpace( $sf ) ),
		 ( $self->numSFEntries( $sf ) /
		   $self->sfHitSpace( $sf ) ) );
    }
}

=head2 idToSF

  Title:    idToSF
  Function: accessor
  Returns:  superfamily identifier for a given astral identifier
  Args:     astral identifier( for example, 'd1dlwa_' )

=cut

sub idToSF {
    my $self = shift;
    my $asid = shift;
    return $self->{ 'id2sf' }->{ $asid };
}

=head2 covSizes

  Title:    covSizes
  Function: accessor
  Returns:  3 element array of size of 'coverage' space for each
    family normalization method
  Args:     none

=cut

sub covSizes {
    my $self = shift;
    my $sf;
    if ( $self->{ 'covSizes' } ) {
	return @{ $self->{ 'covSizes' } };
    }
    else {
	my @covSizes = ( 0, 0, 0 );
	foreach $sf ( $self->sfIds() ) {
	    unless ( $self->numSFUniq( $sf ) == 1 ) {
		$covSizes[ 0 ] += $self->sfHitSpace( $sf );
		$covSizes[ 1 ] += 1;
		$covSizes[ 2 ] += $self->numSFEntries( $sf );
	    }
	}
	$self->{ 'covSizes' } = [ @covSizes ];
	return @{ $self->{ 'covSizes' } };
    }
}

=head2 totalEntries

  Title:    totalEntries
  Function: accessor
  Returns:  scalar number of entries in this astral database.
    Same as sum of size of all families.
  Args:     none

=cut

sub totalEntries {
    my $self = shift;
    return $self->{ 'totalEntries' };
}

=head2 totalSFs

  Title:    totalSFs
  Function: accessor
  Returns:  scalar total number of superfamilies 
  Args:     none

=cut

sub totalSFs {
    my $self = shift;
    return $self->{ 'totalSFs' };
}

=head2 _parseInput

  Title:    _parseInput
  Function: internal
  Returns:  object data structre
  Args:     astral filename, (restriction filename - optional)  If a
    restriction file is specified then superfamilies are restricted to
    only those sequences that mentioned in the restriction file

=cut

sub _parseInput {
    my $astral_file = shift;
    my $rest_file   = shift;

    my ( %SFs, %self, %id2sf, %astral_ids );
    my ( @elements, @keys );
    my ( $l, $id, $sf, $entries, $totalSFs, $qid, $qid_pnt );

    # Parse astral database
    open( AST, $astral_file ) or 
	die( "can not parse $astral_file: $!" );
    while( $l = <AST> ) {
	if ( $l =~ /^>/ ) {
	    my @elements = split( /\s+/, $l );
	    ($id) = ($elements[0] =~ /^>(.+)/);
	    ($sf) = ($elements[1] =~ /^(\w+\.\d+\.\d+)\.\d+/);
	    $id2sf{ $id } = $sf;
	    $astral_ids{ $id } += 1;
	    $SFs{ $sf } = [ &_updateSF( 1, $SFs{ $sf } ) ]; 
            # all entries should be
            # present once when we parse the astral database
	    $entries++;
	}
    }
    close( AST ) or die( "$astral_file: $!" );
    $self{ 'astral_ids' } = \%astral_ids;

    # if a sasum file was specified, then restrict everything to what
    # is mentioned there
    if ( $rest_file ) {
	%SFs = ();
	$entries == 0;
	$qid_pnt = &_parseRestrict( $rest_file );
	foreach $qid ( keys %{ $qid_pnt } ) {
	    $SFs{ $id2sf{ $qid } } =
		[ &_updateSF( $qid_pnt->{ $qid }, 
			      $SFs{ $sf } ) ];
	    # update sizes of this sf with info about this entry.
	    # This entry may be present multiple times
	    $entries += $qid_pnt->{ $qid };
	}
	$self{ 'astral_ids' } = $qid_pnt;
    }

    @keys = keys %SFs;
    $totalSFs = $#keys + 1;
    $self{ 'SFs' } = \%SFs;
    $self{ 'totalEntries' } = $entries;
    $self{ 'totalSFs' } = $totalSFs;
    $self{ 'id2sf' } = \%id2sf;
    
    return \%self;
}

=head2 restrict

  Title:    restrict
  Function: mutator
  Returns:  nothing
  Args:     filename with list of ids to restrict to

=cut

sub restrict {
    my $self    = shift;
    my $rest_fn = shift;
    my %SFs;
    my $entries = 0;
    my ( $qid, $totalSFs, $qid_pnt );
    my ( @query_ids, @keys );

    $qid_pnt = &_parseRestrict( $rest_fn );
    foreach $qid ( keys %{ $qid_pnt } ) {
	$SFs{ $self->idToSF( $qid ) } = 	
	    [ &_updateSF( $self,
			  $qid_pnt->{ $qid }, 
			  $SFs{ $self->idToSF( $qid ) } ) ];
	$entries += $qid_pnt->{ $qid };
    }    
    $self->{ 'astral_ids' } = $qid_pnt;
    
    @keys = keys %SFs;
    $totalSFs = $#keys + 1;

    $self->{ 'SFs' } = \%SFs;
    $self->{ 'totalEntries' } = $entries;
    $self->{ 'totalSFs' } = $totalSFs;
    $self->{ 'covSizes' } = ();
    
}

=head2 _updateSF

  Title:    _updateSF
  Function: internal
  Returns:  array of sizes of this sf
  Args:     how many occurances of this entry to add to sf sizes,
    id of sf to update

=cut

sub _updateSF {
    my ( $self, $n, $sf_pnt );
    my @new_sizes;
    if ( $#_ == 2 ) {
	# we've been blessed and self is defined
	$self   = shift;
	$n      = shift;
	$sf_pnt = shift;
    }
    else {
	# doing all this for the first time
	$n      = shift;
	$sf_pnt = shift;
    }

    if ( $sf_pnt ) {
	# we've seen this sf before
	$new_sizes[ 0 ] = $sf_pnt->[ 0 ] + $n;
	$new_sizes[ 1 ] = $sf_pnt->[ 1 ] + 
	    ( 2 * $n * $sf_pnt->[ 0 ] );
	$new_sizes[ 2 ] = $sf_pnt->[ 2 ] + 1;
    }

    else {
	# first time for this sf
	$new_sizes[ 0 ] = $n;
	$new_sizes[ 1 ] = 0;
	$new_sizes[ 2 ] = 1;
    }

    return @new_sizes;
}

=head2 _parseRestrict

  Title: _parseRestrict
  Function: internal
  Returns: array of query ids in a sasum file
  Args: sasum filename

=cut

sub _parseRestrict {
    my %query_ids;
    my $l1;
    my $restrict_fn = shift;

    open( REST, $restrict_fn ) or 
	die( "can not open $restrict_fn: $!" );
    
    while ( chomp( $l1 = <REST> ) ) {
	$query_ids{ $l1 } += 1;
    }

    close( REST ) or die( "could not close $restrict_fn: $!" );
    
    return ( \%query_ids );
}
1;
