#!/usr/bin/env perl

# (c) 2002 Ed Green, UC Regents

use strict;
use Getopt::Std;
use FamNorm;
use Bio::SearchIO;
use Carp;

use vars qw( $opt_q $opt_t $opt_r $opt_d $opt_a $opt_f );

# This string is used as the argument for the 
# FORMAT of the bioperl parser, if applicable
my %FORMATS = ( 'B' => 'blast',
		'P' => 'psiblast',
		'F' => 'fasta',
		'D' => 'doublet_align',
		'W' => 'wu-blast', 
		'G' => 'g-search' );

my %done; # a hash of all the query ids that we see.  We will compare
	  # this to all the query ids in the query database to make
	  # sure our results are complete, if our databases are astral
	  # databases

my %astral_ids; # An hash of the ids in the astral database we're
		# using 

MAIN : {
    
    my ( $q_fno, $t_fno );

    &init();

    &outputHeader();

    if ( $opt_a ) {
	( $q_fno, $t_fno ) = &dbStats();
	&setUpAstralIds( $q_fno->astralIds() );
    }

    &outputSeparator();

    # Look in the directory specified
    if ( -d $opt_d ) {
	&processThisDir( $opt_d, $q_fno, $t_fno );
    }
    elsif ( &hasRestriction( $opt_d ) ) {
	&processThisResult( "$opt_d", 
			    $q_fno, $t_fno );
    }

    if ( $opt_a ) {
	&compCheck();
    }

}

sub processThisDir {
    my $dir   = shift;
    my $q_fno = shift;
    my $t_fno = shift;

    my ( $file_or_dir );

    opendir( D, $dir ) or die( "Can not open $dir: $!" );
    my @in_dir = readdir( D );
    closedir( D );

    foreach $file_or_dir ( @in_dir ) {
	if ( -d "$dir/$file_or_dir" && $file_or_dir ne '.' &&
	     $file_or_dir ne '..' ) {
	    &processThisDir( "$dir/$file_or_dir" );
	}
	elsif ( &hasRestriction( $file_or_dir ) ) {
	    &processThisResult( "$dir/$file_or_dir", 
				$q_fno, $t_fno );
	}
	else {
	    # wasn't a proper results file or a directory
	}
    }

}

sub processThisResult {
    my $res_file = shift;
    my $q_fno    = shift;
    my $t_fno    = shift;
    my $comp;

    my ( $searchio, $result, $qid, $hit, $tid, $sig );

    if ( $opt_f eq 'D' ) {
	# doublet_align format, can't use Bioperl parser (yet)
	&processDoublet( $res_file, $q_fno, $t_fno );
    }

    elsif ( $opt_f eq 'F' ) {
	# fasta format - don't use the bioperl parser...it's buggy!
	&processFasta( $res_file, $q_fno, $t_fno );
    }

    elsif ( $opt_f eq 'W' ) {
	# wu_blast - don't use the bioperl parser ... it's buggy!
	&processWuBlast( $res_file, $q_fno, $t_fno );
    }

    elsif ( $opt_f eq 'G' ) {
	&processGSearch( $res_file, $q_fno, $t_fno );
    }

    else {
	# one of the other formats -- use the bioperl parser
	if ( $comp = &isCompressed( $res_file ) ) {
	    open( FH, "$comp"."cat $res_file|" ) 
		or die( "$comp"."cat $res_file: $!" );
	    # Make a Bio::Search::Result::ResultI object for each file
	    $searchio = 
		new Bio::SearchIO( -format => $FORMATS{ $opt_f },
				   -fh   => \*FH );
	}
	else {
	    $searchio = 
		new Bio::SearchIO( -format => $FORMATS{ $opt_f },
				   -file   => $res_file );
	}
	
	while ( $result = $searchio->next_result() ) {
	    $qid = &findQid( $res_file );
	    
	    unless ( &checkQid( $qid ) ) {
		carp( "Resolved file $res_file to astral id $qid\n",
		      "which is not in the astral database $opt_a\n" );
	    }	    
	    
	    while( $hit = $result->next_hit() ) {
		$tid = $hit->name();
		$sig = $hit->significance();
		unless ( $tid eq $qid ) {
		    &outputLine( $qid, $tid, $sig, 
				 &findRel( $qid, $tid, 
					   $q_fno, $t_fno ),
				 &normalize( $qid, $tid,
					     $q_fno, $t_fno ) );
		}
	    }
	}
	close( FH );
    }
}

sub processDoublet {
    my $fn = shift;
    my $q_fno = shift;
    my $t_fno = shift;
    my ( $qid, $tid, $sig, $comp );

    my ( $q_len, $lambda, $K, $H, $t_len, $raw );

    my $found_query_line = 0;
    my $some_results = 0;
    my @el;

    if ( $comp = &isCompressed( $fn ) ) {
	open( FH, "$comp"."cat $fn|" ) 
		or die( "$comp"."cat $fn: $!" );
    }
    else {
	open( FH, "$fn" ) or die( "$fn: $!" );
    }


    chomp( $_ = <FH> );

    until( $found_query_line ) {
	unless ( !($_) || ($_ =~ /^\#/) ) {
	    ( $qid, $q_len, $lambda, $K, $H ) = split( /\s+/, $_ );
	    $found_query_line = 1;
	}
	chomp( $_ = <FH> );
    }

    
    #now we're in the stats section, every line is a hit
    #                         tid    (s-w) z-score ?? E
    while( chomp( $_ = <FH> ) ) {
	unless ( !($_) || ($_ =~ /^\#/) ) {
	    ($tid, $t_len, $raw, $sig ) = split( /\s+/, $_ );
	    unless ( $tid eq $qid ) {
		$some_results = 1;
		&outputLine( $qid, $tid, $sig,
			     &findRel( $qid, $tid,
				       $q_fno, $t_fno ),
			     &normalize( $qid, $tid,
					 $q_fno, $t_fno ) );
	    }
	}
    }

    close (FH);
    unless ( $some_results ) {
	carp( "No hits found in $fn for $qid" );
    }
}

sub processFasta {
    my $fn = shift;
    my $q_fno = shift;
    my $t_fno = shift;
    my ( $qid, $tid, $sig, $comp );

    my $found_qid = 0;
    my $found_stat_section = 0;
    $qid = &findQid( $fn );

    unless ( &checkQid( $qid ) ) {
	carp( "Resolved file $fn to astral id $qid\n",
	      "which is not in the astral database $opt_a\n" );
    }
    
    if ( $comp = &isCompressed( $fn ) ) {
	open( FH, "$comp"."cat $fn|" ) 
	    or die( "$comp"."cat $fn: $!" );
    }
    else {
	open( FH, "$fn" ) or die( "$fn: $!" );
    }
    
    $_ = <FH>;
    until ( $found_stat_section ) {
	if ( $_ = <FH> ) {
	    if ( /^The best scores are:/ ) {
		$found_stat_section = 1;
	    }
	}
	else {
	    return;
	}
    }

    #now we're in the stats section, every line is a hit
    #                         tid    (s-w) z-score ?? E
    $_ = <FH>;
    while( ($tid, $sig) = /^(\S+?)\s+.+?(\S+)\Z/ ) {
	unless ( $tid eq $qid ) {
	    &outputLine( $qid, $tid, $sig,
			 &findRel( $qid, $tid,
				   $q_fno, $t_fno ),
			 &normalize( $qid, $tid,
				     $q_fno, $t_fno ) );
	}
	$_ = <FH>;
    }
    close (FH);
}

sub processWuBlast {
    my $fn = shift;
    my $q_fno = shift;
    my $t_fno = shift;
    my ( $qid, $tid, $sig, $comp );

    my $found_qid = 0;
    my $found_stat_section = 0;
    my $finished_res = 0;
    $qid = &findQid( $fn );

    unless ( &checkQid( $qid ) ) {
	carp( "Resolved file $fn to astral id $qid\n",
	      "which is not in the astral database $opt_a\n" );
    }
    
    if ( $comp = &isCompressed( $fn ) ) {
	open( FH, "$comp"."cat $fn|" ) 
	    or die( "$comp"."cat $fn: $!" );
    }
    else {
	open( FH, "$fn" ) or die( "$fn: $!" );
    }
    
    $_ = <FH>;
    until ( $found_stat_section ) {
	if ( $_ = <FH> ) {
	    if ( /^Sequences producing High-scoring Segment Pairs:/ ) {
		$found_stat_section = 1;
	    }
	    elsif ( /^WARNING:  No precomputed values / ) {
#		carp( "Wu-blast doesn't like this matrix\n" );
	    }
	}
	else {
	    return;
	}
    }

    #now we're in the stats section, every line is a hit
    #                         tid    (s-w) z-score ?? E
    $_ = <FH>; # skip blank line
    while ( $_ = <FH> ) {
	if ( $_ =~ /^\n$/ ) {
	    $finished_res = 1;
	}
	if ( $finished_res ) {
	    close( FH );
	    return;
	}
	else {
	    $tid = substr( $_,  0, 7 );
	    $sig = substr( $_, 68, 9 );
	    $sig = &removeWhiteSpace( $sig );
	    unless ( $tid eq $qid ) {
		&outputLine( $qid, $tid, $sig,
			     &findRel( $qid, $tid,
				       $q_fno, $t_fno ),
			     &normalize( $qid, $tid,
					 $q_fno, $t_fno ) );
	    }
	}
    }
}

sub processGSearch {
    my $fn = shift;
    my $q_fno = shift;
    my $t_fno = shift;
    my ( $qid, $tid, $sig, $comp );

    my $found_qid = 0;
    my $found_stat_section = 0;
    $qid = &findQid( $fn );

    unless ( &checkQid( $qid ) ) {
	carp( "Resolved file $fn to astral id $qid\n",
	      "which is not in the astral database $opt_a\n" );
    }
    
    if ( $comp = &isCompressed( $fn ) ) {
	open( FH, "$comp"."cat $fn|" ) 
	    or die( "$comp"."cat $fn: $!" );
    }
    else {
	open( FH, "$fn" ) or die( "$fn: $!" );
    }
    
    $_ = <FH>;
    until ( $found_stat_section ) {
	if ( $_ = <FH> ) {
	    if ( /^Search space size/ ) {
		$found_stat_section = 1;
	    }
	}
	else {
	    return;
	}
    }
    
    $_ = <FH>; # skip the blank line
    #now we're in the stats section, every line is a hit
    #                         tid    (s-w) z-score ?? E
    while( $_ = <FH> ) {
	$sig = substr( $_, 16, 9 );
	$tid = substr( $_, 26, 7 );

	$sig = &removeWhiteSpace( $sig );

	unless ( $tid eq $qid ) {
	    &outputLine( $qid, $tid, $sig,
			 &findRel( $qid, $tid,
				   $q_fno, $t_fno ),
			 &normalize( $qid, $tid,
				     $q_fno, $t_fno ) );
	}
    }
    close (FH);
}

sub removeWhiteSpace {
    # takes any leading or trailing whitespace off of the argument
    # and returns the argument
    my $arg = shift;
    $arg =~ s/^\s+//;
    $arg =~ s/\s+$//;
    return $arg;
}

sub findQid {
    # find what the id of the query sequence is based on the filename
    # of output file; also, add an entry to the global %done for this
    # query 
    my $path_fn = shift;
    my $fn;
    my $qid;

    # if the filename has a bunch of path stuff (usually the case),
    # just concentrate on the last part, after the final '/'
    if ( $path_fn =~ /\/.+/ ) {
	($fn) = ( $path_fn =~ /.+\/(.+)/ );
    }
    else {
	$fn = $path_fn;
    }

    $qid = substr( $fn, 0, 7 );
    
    #now, the id we got darn well better have 7 characters; we could
    #do more to check that it's sensible...maybe later
    unless ( length( $qid ) == 7 ) {
	carp( "Problem getting query id from filename\n".
	      "Got: $qid from $path_fn\n" );
    }
    
    $done{ $qid } = 1;

    return $qid;
    
}

sub checkQid {
    my $qid = shift;
    
    if ( $astral_ids{ $qid } ) {
	# found it
	return 1;
    }

    else {
	# oops, not here
	return 0;
    }
}
    

sub hasRestriction {
    my $filename = shift;
    if ( index( $filename, $opt_r ) >= 0 ) {
	return 1;
    }
    return 0;
}

sub isCompressed {
    my $fn = shift;
    if ( substr( $fn, length( $fn ) - 2 ) eq 'gz' ) {
	return 'gz';
    }
    elsif ( substr( $fn, length( $fn ) - 3 ) eq 'bz2' ) {
	return 'bz';
    }
    else {
	return 0;
    }
}

sub outputLine {
    my $qid = shift;
    my $tid = shift;
    my $sig = shift;
    my $rel = shift;
    my $unorm = shift;
    my $qnorm = shift;
    my $lnorm = shift;
    print( "$qid $tid $sig $rel $unorm $lnorm $qnorm\n" );
}

sub findRel {
    my $qid = shift;
    my $tid = shift;
    my $q_fno = shift;
    my $t_fno = shift;
    
    my $qsf = $q_fno->idToSF( $qid );
    my $tsf = $t_fno->idToSF( $tid );

    my @qsf = split( /\./, $qsf );
    my @tsf = split( /\./, $tsf );
    
    my $rel = 0;
    while( $qsf[ $rel ] eq $tsf[ $rel ] && 
	   $rel < 3 ) {
	$rel++;
    }
    return $rel;
}

sub normalize {
    my $qid   = shift;
    my $tid   = shift;
    my $q_fno = shift;
    my $t_fno = shift;

    # Query and target databases are the same
    if ( $q_fno == $t_fno ) {
	return ( $q_fno->normFactor( $q_fno->idToSF( $qid ) ) );
    }
    else {
	return ( 1, 
		 $q_fno->numSFEntries($q_fno->idToSF($qid)) );
    }
}

sub outputHeader {
    print( "dbvdbsum v1.0 results\n" );
    print( "Query database: $opt_q\n" );
    print( "Target database: $opt_t\n" );
}

sub dbStats {
    my ( $q_fno, $t_fno );
    $q_fno = new FamNorm( $opt_q );
    if ( $opt_q eq $opt_t ) {
	$t_fno = $q_fno;
    }
    else {
	$t_fno = new FamNorm( $opt_t );
    }
    my @q_rels = $q_fno->covSizes();
    my @t_rels = $t_fno->covSizes();
    
    print( "Query database size: ",
	   $q_fno->totalEntries(), "\n" );
    print( "Target database size: ",
	   $t_fno->totalEntries(), "\n" );
    print( "Query superfamily relations: ",
	   $q_rels[ 0 ], " ", $q_rels[ 2 ], " ",
	   $q_rels[ 1 ], "\n" );
    print( "Query superfamily relations: ",
	   $t_rels[ 0 ], " ", $t_rels[ 2 ], " ",
	   $t_rels[ 1 ], "\n" );
    return ( $q_fno, $t_fno );
}

sub outputSeparator {
    print( "query_id target_id score scop_rel unnorm lin_norm quad_norm\n" );
    print( "__HITS__\n" );
}

sub compCheck {
    my $qid;
    foreach $qid ( keys %astral_ids ) {
	if ( $done{ $qid } == 1 ) {
	    $done{ $qid } = 0; # seen it, flip it
	}
	else {
	    carp( "Did not see any result for $qid" );
	}
    }
    
    foreach $qid ( @_ ) {
	unless ( $done{ $qid } == 0 ) {
	    carp( "$qid seen, but it isn't in the query db" );
	}
    }
}

sub setUpAstralIds {
    foreach my $id ( @_ ) {
	$astral_ids{ $id } = 1;
    }
}

sub init {
    getopts( 'q:t:r:d:a:f:' );
    unless( $opt_a && $opt_f ) {
	print( "dbvdbsum.pl Database vs database search results summarizer\n" );
	print( "Examines the output files of a database vs\n" );
	print( "database search and generates a summary file\n" );
	print( "describing the reported hits.\n" );
	print( "Do \$perldoc -m dbvdbsum.pl for a more complete\n" );
	print( "description.\n" );
	print( "OPTIONS:\n" );
	print( "-q [ query database ]\n" );
	print( "-t [ target database; Default = same as -q ]\n" );
	print( "-r [ filename restriction; Default = ]\n" );
	print( "     none -- look at all files ]\n" );
	print( "-d [ directory to examine for results files;\n" );
	print( "      Default = . ]\n" );
	print( "-a [ true value means databases are astral\n" );
	print( "     databases, so show some stats about them ]\n" );
	print( "-f [ format of results; (B, F, P, D, W) ]\n" );
	exit ( 0 );
    }
    
    $opt_t = $opt_q unless $opt_t;
    $opt_r = "" unless $opt_r;
    $opt_d = '.' unless $opt_d;
    $opt_a = 1 unless $opt_a;
}
