#!/usr/bin/env perl

# (c) 2002, Ed Green / UC Regents

# sfSizePerf.pl
# Takes an astral database and a dbvdbsum file of search results of
# searching that database against itself and generates data showing
# how performance differs as superfamilies grow.  The output file
# looks like this (data is for illustration only):
#
# # Superfamily performance distribution for [INPUT DBVDB] and [INPUT
# ASTRAL BD] at [X] EPQ
# # Size     Percenct found at [X] EPQ
# 1          0.2
# 2          0.223
# 3          0.24
# .          .
# .          .
# .          .
# [LARGEST]  0.51

use strict;
use FamNorm;
use DBVDB;
use Getopt::Std;
use vars qw( $opt_a $opt_s $opt_e );

MAIN : {

    my ( 
	 $fno, 
	 $sso, 
	 $queries,
	 $sc,
	 $this_hit,
	 $errors,
	 $sf,
	 $size,
	 $cov,
	 $inc,
	 $und,
	 $total_size,
	 $hits_p
	 );

    my ( 
	 @size_cov,
	 @sizes,
	 @sfs
	 );

    my ( 
	 %sf_cov
	 );

    &init();
    
    # set up FamNorm object, $fno
    $fno = FamNorm->new( $opt_a );

    # set up Sasum object, $sso
    $sso = DBVDB->new( $opt_s );

    $queries = $sso->numQueries();

    # go from best score to worst score until we hit the cutoff EPQ.
    # At each score, update the %sf_cov, which should look like this:
    # %sf_cov = { SF_NAME => [ # CORRECT, # INCORRECT, # UNDETERMINED ],
    #                         ...
    #             SF_NAME => [ # CORRECT, # INCORRECT, # UNDETERMINED ] }
    foreach $sc ( $sso->sigs() ) {
	$hits_p = $sso->hitsAtSig( $sc );
	
	foreach $this_hit ( @{ $hits_p } ) {
	    $errors += &updateSFCov( $this_hit, $fno, $sso, 
				     \%sf_cov, $queries ); 
	}

	last if $errors/$queries  >= $opt_e;
    }

    # make @sizes that has an anonymous array for each element that
    # contains the SF ids of all SFs of the size of the index of the
    # array position
    @sfs = $fno->sfIds();
    foreach $sf ( @sfs ) {
	$size = $fno->numSFUniq( $sf );
	push( @{ $sizes[ $size ] }, $sf );
    }

    # Show how each superfamily did, in order of size
    print ( "# Superfamily Coverages\n" );
    print ( "# ID   NUM_CORRECT  NUM_INCORRECT  NUM_UNDETERMINED\n" );      
    for ( $size = 0; $size <= $#sizes; $size++ ) {
	if ( $sizes[ $size ] ) {
	    print( "# SIZE $size\n" );
	    foreach $sf ( @{ $sizes[ $size ] } ) {
		print( "# $sf " );
		print( &orZero( $sf_cov{ $sf }->[ 0 ] ), " " );
		print( &orZero( $sf_cov{ $sf }->[ 1 ] ), " " );
		print( &orZero( $sf_cov{ $sf }->[ 2 ] ), "\n" );
	    }
	}
    }

    # Show how all SFs of a given size did, cummulatively
    print( "# Size vs coverage of all SF's at that size\n" );
    
    for ( $size = 1; $size <= $#sizes; $size++ ) {
	$cov = 0;
	$inc = 0;
	$und = 0;
	$total_size = 0;
	foreach $sf ( @{ $sizes[ $size ] } ) {
	    $cov += $sf_cov{ $sf }->[ 0 ];
	    $inc += $sf_cov{ $sf }->[ 1 ];
	    $und += $sf_cov{ $sf }->[ 2 ];
	    $total_size += ( ($size * $size) - $size );
	}
	if ( $total_size == 0 ) { # none at this size or size == 1
	    if ( $size == 1 ) {
		print( "$size 0 $inc $und\n" );
	    }
	    else {
		print( "$size 0 0 0\n" );
	    }
	}
	else {
	    print( "$size ", $cov / $total_size, " $inc $und\n" );
	}
    }
}

sub orZero {
    my $str = shift;
    if ( defined( $str ) ) {
	return $str;
    }
    else {
	return 0;
    }
}

sub updateSFCov {
    my $hit = shift;
    my $fno = shift;
    my $sso = shift;
    my $sf_cov_pnt = shift;
    my $queries = shift;

    # find out what SF this hit is in
    my $sf = $fno->idToSF( $sso->hitQueryId( $hit ) );

    # find out if it's correct, incorrect, or undetermined
    my $rel = $sso->hitRel( $hit );
    # $sso->hitRel( $hit->[ 0 ], $hit->[ 1 ] );

    # update the correct element of the correct sf in $sf_cov_pnt
    if ( $rel >= 3 ) {
	$sf_cov_pnt->{ $sf }->[ 0 ]++;
	return 0;
    }
    elsif ( $rel == 2 ) {
	$sf_cov_pnt->{ $sf }->[ 2 ]++;
	return 0;
    }
    else {
	$sf_cov_pnt->{ $sf }->[ 1 ]++;
	return 1;
    }
}

sub init {
    getopts( 's:a:e:' );
    unless ( $opt_a && $opt_a ) {
	print( "sfSizePerf.pl\n" );
	print( "Shows how much coverage was generated for each\n" );
	print( "superfamily, sorted by size.  Also generates SF\n" );
	print( "size vs total coverae at that size distribution,\n" );
	print( "suitable for gnuplot.\n" );
	print( "OPTIONS:\n" );
	print( "-a [ astral database used ]\n" );
	print( "-s [ dbvdbsum file with results ]\n" );
	print( "-e [ EPQ cutoff; default = 0.01 ]\n" );
	exit( 0 );
    }

    unless ( $opt_e ) {
	$opt_e = 0.01;
    }
}
